import os
import btllib
import unittest


class SeqReaderTests(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(__file__)

    def test_seq_reader_single_seq(self):
        seq = 'CGCGTGAAAGCAAAACAAGA'
        path = os.path.join(self.base_dir, 'single_seq.fa')
        with btllib.SeqReader(path, btllib.SeqReaderFlag.SHORT_MODE) as reader:
            read = [record.seq for record in reader]
        self.assertListEqual([seq], read)
        
    def test_fasta_reader(self):
        ids = ["asdf", "ghjk"]
        seqs = ["ACTG", "TGCA"]

        for iteration in range(3):
            reader = btllib.SeqReader(os.path.join(self.base_dir, "../input.fa.gz.bz2.xz"),
                                      btllib.SeqReaderFlag.SHORT_MODE)
            self.assertEqual(reader.get_format(), btllib.SeqReader.SeqReaderFormat_FASTA)

            i = 0
            for record in reader:
                self.assertEqual(record.id, ids[i])
                self.assertEqual(record.seq, seqs[i])
                self.assertEqual(record.qual, "")
                i += 1
            self.assertEqual(i, 2)
            
    def test_multiline_fasta_reader(self):
        ids = ["asdf", "ghjk"]
        seqs = ["ACTG", "TGCA"]
       
        for iteration in range(3):
            reader = btllib.SeqReader(os.path.join(self.base_dir, "../input_multiline.fa"),
                            btllib.SeqReaderFlag.SHORT_MODE)

            self.assertEqual(reader.get_format(), btllib.SeqReader.SeqReaderFormat_FASTA)

            i = 0
            for record in reader:
                self.assertEqual(record.id, ids[i])
                self.assertEqual(record.seq, seqs[i])
                i += 1

            self.assertEqual(i, 2)
    
    def test_fastq_reader(self):
        ids = ["asdf", "ghjk"]
        seqs = ["ACTG", "TGCA"]
        quals = ["!@^&", "(#&$"]

        for iteration in range(3):
            reader = btllib.SeqReader(os.path.join(self.base_dir, "../input.fq.tar.xz"),
                                      btllib.SeqReaderFlag.SHORT_MODE)
            self.assertEqual(reader.get_format(), btllib.SeqReader.SeqReaderFormat_FASTQ)

            i = 0
            for record in reader:
                self.assertEqual(record.id, ids[i])
                self.assertEqual(record.seq, seqs[i])
                self.assertEqual(record.qual, quals[i])
                i += 1
            self.assertEqual(i, 2)
    
    def test_multiline_fastq_reader(self):
        ids = ["asdf", "ghjk"]
        seqs = ["ACTG", "TGCA"]
        quals = ["!@^&", "(#&$"]
        
        for iteration in range(3):
            reader = btllib.SeqReader(os.path.join(self.base_dir, "../input_multiline.fq"),
                            btllib.SeqReaderFlag.SHORT_MODE)

            self.assertEqual(reader.get_format(), btllib.SeqReader.SeqReaderFormat_FASTQ)

            i = 0
            for record in reader:
                self.assertEqual(record.id, ids[i])
                self.assertEqual(record.seq, seqs[i])
                i += 1

            self.assertEqual(i, 2)
''' 
NOTE: This unit test module is currently passing on all platforms except for the Linux CI test. Although you can use this code as a reference for your implementation, we have commented it out to prevent CI failures. Please be aware of the limitations and potential issues when using this example. To re-enable the test, simply uncomment the block of code below.

    def test_SAM_reader(self):
        ids = ["q1", "q2"]
        seqs = ["ACTG", "TGCA"]
        quals = ["!@^&", "(#&$"]

        for iteration in range(3):
            reader = btllib.SeqReader(os.path.join(self.base_dir, "../input.bam"),
                                      btllib.SeqReaderFlag.SHORT_MODE)
            self.assertEqual(reader.get_format(), btllib.SeqReader.SeqReaderFormat_SAM)

            i = 0
            for record in reader:
                self.assertEqual(record.id, ids[i])
                self.assertEqual(record.seq, seqs[i])
                self.assertEqual(record.qual, quals[i])
                i += 1
            self.assertEqual(i, 2)
'''



import os
import unittest
import btllib

class TestCountingBloomFilter8(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(__file__)

    def test_insert_and_contains(self):
        cbf = btllib.CountingBloomFilter8(1024 * 1024, 3)
        element = ([1, 10, 100])
        cbf.insert(element)
        cbf.insert(element)
        self.assertEqual(cbf.contains(element), 2)

    def test_absent_element(self):
        cbf = btllib.CountingBloomFilter8(1024 * 1024, 3)
        element = ([1, 10, 100])
        cbf.insert(element)
        cbf.insert(element)

        absent_element = ([1, 20, 100])
        self.assertEqual(cbf.contains(absent_element), 0)

    def test_save_load(self):
        cbf = btllib.CountingBloomFilter8(1024 * 1024, 3)
        element = ([1, 10, 100])
        cbf.insert(element)
        cbf.insert(element)
        
        file_path = os.path.join(self.base_dir, "test.cbf")
        cbf.save(file_path)
        try:
            loaded_cbf = btllib.CountingBloomFilter8(file_path)
            element = ([1, 10, 100])
            self.assertEqual(loaded_cbf.contains(element), 2)
        finally:
            os.remove(file_path)
    
    def test_kmer_counting_bloom_filter(self): 
        seq = "CACTATCGACGATCATTCGAGCATCAGCGACTG"
        seq2 = "GTAGTACGATCAGCGACTATCGAGCTACGAGCA"
        
        self.assertEqual(len(seq), len(seq2))
       
        kmer_cbf = btllib.KmerCountingBloomFilter8(1024 * 1024, 4, len(seq) // 2)
        
        kmer_cbf.insert(seq)
        expected = len(seq) - len(seq) // 2 + 1
        
        self.assertEqual(kmer_cbf.contains(seq), expected)
        self.assertLess(kmer_cbf.contains(seq2), 1)



import os
import btllib
import re
import unittest


class RandSeqTests(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(__file__)

    def test_random_sequence_dna(self):
        dna_re = re.compile(r"^[ACTG]*$")
        seq_generator = btllib.RandSeq(btllib.RandSeq.SeqType_DNA)
        random_seq = seq_generator.generate(100)
        self.assertEqual(len(random_seq), 100)
        self.assertTrue(re.search(dna_re, random_seq))

    def test_random_sequence_rna(self):
        rna_re = re.compile(r"^[ACUG]*$")
        seq_generator = btllib.RandSeq(btllib.RandSeq.SeqType_RNA)
        random_seq = seq_generator.generate(200)
        self.assertEqual(len(random_seq), 200)
        self.assertTrue(re.search(rna_re, random_seq))



import os
import btllib
import unittest


class IndexlrTests(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(__file__)

    def test_indexlr_single_seq(self):
        minimizers = [9638000835411166970, 16169464608842636080, 15783021188953468643]
        seq_id = "1"
        path = os.path.join(self.base_dir, 'single_seq.fa')
        with btllib.Indexlr(path, 10, 5, btllib.IndexlrFlag.LONG_MODE) as indexlr:
            for record in indexlr:
                self.assertEqual(seq_id, record.id)
                self.assertEqual(minimizers, [mx.out_hash for mx in record.minimizers])



from setuptools import setup, find_packages

setup(name='btllib',
      version='@PROJECT_VERSION@',
      packages=find_packages(include=['btllib']),
      package_data={'btllib': ['_btllib.so']},
      url='https://github.com/bcgsc/btllib')



import os
import string
import btllib
import unittest

class BloomFilterTests(unittest.TestCase):
    
    def setUp(self):
        self.base_dir = os.path.dirname(__file__)
    
    def test_bloom_filter(self):
        bf = btllib.BloomFilter(1024 * 1024, 3, "ntHash")
        bf.insert([1, 10, 100])
        bf.insert([100, 200, 300])

        self.assertTrue(bf.contains([1, 10, 100]))
        self.assertTrue(bf.contains([100, 200, 300]))
        self.assertFalse(bf.contains([1, 20, 100]))

    def test_kmer_bloom_filter(self):
        seq = "CACTATCGACGATCATTCGAGCATCAGCGACTG"
        seq2 = "GTAGTACGATCAGCGACTATCGAGCTACGAGCA"

        kmer_bf = btllib.KmerBloomFilter(1024 * 1024, 4, len(seq) // 2)
        kmer_bf.insert(seq)
        self.assertEqual(kmer_bf.contains(seq), len(seq) - len(seq) // 2 + 1)
        self.assertLessEqual(kmer_bf.contains(seq2), 1)

    def test_seed_bloom_filter(self):
        seed1 = "000001111111111111111111111111111"
        seed2 = "111111111111111111111111111100000"
        seq = "CACTATCGACGATCATTCGAGCATCAGCGACTG"
        snp_seq1 = "AACTATCGACGATCATTCGAGCATCAGCGACTG"
        snp_seq2 = "CACTATCGACGATCATTCGAGCATCAGCGACTA"

        seed_bf = btllib.SeedBloomFilter(1024 * 1024, len(seq), [seed1, seed2], 4)
        seed_bf.insert(seq)
        hit_seeds = seed_bf.contains(seq)

        self.assertIn(0, hit_seeds[0])
        self.assertIn(1, hit_seeds[0])

        hit_seeds = seed_bf.contains(snp_seq1)
        self.assertIn(0, hit_seeds[0])
        self.assertNotIn(1, hit_seeds[0])

        hit_seeds = seed_bf.contains(snp_seq2)
        self.assertNotIn(0, hit_seeds[0])
        self.assertIn(1, hit_seeds[0])
        
    def test_bloom_filter_save_load(self):
        bf = btllib.BloomFilter(1024 * 1024, 3, "ntHash")
        bf.insert([1, 10, 100])
        bf.insert([100, 200, 300])

        file_path = os.path.join(self.base_dir, "test.bf")
        bf.save(file_path)

        try:
            bf2 = btllib.BloomFilter(file_path)

            self.assertEqual(bf2.get_hash_fn(), "ntHash")

            self.assertTrue(bf2.contains([1, 10, 100]))
            self.assertTrue(bf2.contains([100, 200, 300]))
            self.assertFalse(bf2.contains([1, 20, 100]))
        finally:
            os.remove(file_path)




import btllib
import unittest


class CalcPhredAvgTests(unittest.TestCase):

    def test_calc_phred_avg(self):
        qual = "$$%%)*0)'%%&$$%&$&'''*)(((((()55561--.12356577-++**++,////.*))((()+))**010/..--+**++*+++)++++78883"
        self.assertAlmostEqual(
            btllib.calc_phred_avg(qual, 0, 10), 5.34264, places=3)
        self.assertAlmostEqual(btllib.calc_phred_avg(qual), 8.54327, places=3)
        self.assertAlmostEqual(
            btllib.calc_phred_avg(qual, 0, 4), 3.47128, places=3)
        self.assertAlmostEqual(
            btllib.calc_phred_avg(qual, 5, 20), 5.48923, places=3)



import unittest
import btllib

class TestNtHash(unittest.TestCase):

    def test_kmer_hash_values(self):
        seq = "ACATGCATGCA"
        k = 5
        h = 3

        hashes = [
            [0x38cc00f940aebdae, 0xab7e1b110e086fc6, 0x11a1818bcfdd553],
            [0x603a48c5a11c794a, 0xe66016e61816b9c4, 0xc5b13cb146996ffe]
        ]

        nthash = btllib.NtHash(seq, h, k)
        nthash.roll()
        ntblind = btllib.BlindNtHash(seq, h, k)

        for expected_hashes in hashes:
            nthash.roll()
            ntblind.roll(seq[ntblind.get_pos() + 1])
            self.assertEqual(expected_hashes, list(nthash.hashes()))
            self.assertEqual(expected_hashes, list(ntblind.hashes()))

    def test_kmer_rolling(self):
        seq = "AGTCAGTC"
        h = 3
        k = 4

        nthash = btllib.NtHash(seq, h, k)
        hashes = []

        while nthash.roll():
            h_vals = list(nthash.hashes())
            hashes.append(h_vals)

        self.assertEqual(len(hashes), len(seq) - k + 1)
        self.assertEqual(hashes[0], hashes[-1])

    def test_kmer_rolling_vs_ntbase_hash_values(self):
        seq = "ACGTACACTGGACTGAGTCT"
        kmer1 = "ACGTACACTGGACTGAGT"
        kmer2 = "CGTACACTGGACTGAGTC"
        kmer3 = "GTACACTGGACTGAGTCT"

        nthash = btllib.NtHash(seq, 3, len(seq) - 2)
        nthash_vector = [
            btllib.NtHash(kmer1, nthash.get_hash_num(), len(kmer1)),
            btllib.NtHash(kmer2, nthash.get_hash_num(), len(kmer2)),
            btllib.NtHash(kmer3, nthash.get_hash_num(), len(kmer3))
        ]

        i = 0
        while nthash.roll() and nthash_vector[i].roll():
            for j in range(nthash.get_hash_num()):
                self.assertEqual(nthash.hashes()[j], nthash_vector[i].hashes()[j])
            i += 1
        self.assertEqual(i, 3)

    def test_canonical_hashing(self):
        seq_f = "ACGTACACTGGACTGAGTCT"
        seq_r = "AGACTCAGTCCAGTGTACGT"
        h = 3

        nthash_f = btllib.NtHash(seq_f, h, len(seq_f))
        nthash_r = btllib.NtHash(seq_r, h, len(seq_r))

        nthash_f.roll()
        nthash_r.roll()
        self.assertEqual(nthash_f.get_hash_num(), nthash_r.get_hash_num())
        self.assertEqual(list(nthash_f.hashes()), list(nthash_r.hashes()))

    def test_kmer_back_rolling(self):
        seq = "ACTAGCTG"
        h = 3
        k = 5

        nthash = btllib.NtHash(seq, h, k)
        hashes = []

        while nthash.roll():
            h_vals = list(nthash.hashes())
            hashes.append(h_vals)

        self.assertEqual(len(hashes), len(seq) - k + 1)

        while hashes:
            self.assertEqual(list(nthash.hashes()), hashes.pop())
            nthash.roll_back()


    def test_skipping_ns(self):
        seq = "ACGTACACTGGACTGAGTCT"
        seq_with_ns = seq

        self.assertGreaterEqual(len(seq_with_ns), 10)
        middle = len(seq_with_ns) // 2
        seq_with_ns = seq_with_ns[:middle] + "NN" + seq_with_ns[middle + 2:]
        k = (len(seq) - 2) // 2 - 1
        nthash = btllib.NtHash(seq_with_ns, 3, k)

        positions = list(range(0, len(seq_with_ns) // 2 - k + 1)) + \
                    list(range(len(seq_with_ns) // 2 + 2, len(seq_with_ns) - k + 1))

        i = 0
        while nthash.roll():
            self.assertEqual(nthash.get_pos(), positions[i])
            i += 1
        self.assertEqual(len(positions), i)
 
    def test_rna(self):
        seq = "ACGTACACTGGACTGAGTCT"
        dna_nthash = btllib.NtHash(seq, 3, 20)

        rna_seq = "ACGUACACUGGACUGAGUCU"
        rna_nthash = btllib.NtHash(rna_seq, 3, 20)

        dna_nthash.roll()
        rna_nthash.roll()

        for i in range(dna_nthash.get_hash_num()):
            self.assertEqual(dna_nthash.hashes()[i], rna_nthash.hashes()[i])
    
    def test_spaced_seed_hash_values(self):
        seq = "ACATGCATGCA"
        seeds = ["11100111"]
        s_seeds = btllib.parse_seeds(seeds)
        k = len(seeds[0])
        h = 3

        true_hashes = [
            [0x10be4904ad8de5d, 0x3e29e4f4c991628c, 0x3f35c984b13feb20],
            [0x8200a7aa3eaf17c8, 0x344198402f4c2a9c, 0xb6423fe62e69c40c],
            [0x3ce8adcbeaa56532, 0x162e91a4dbedbf11, 0x53173f786a031f45]
        ]

        nthash = btllib.SeedNtHash(seq, s_seeds, h, k)

        for h_vals in true_hashes:
            nthash.roll()
            self.assertEqual(list(nthash.hashes()), h_vals)
            
    def test_spaced_seeds(self):
        seq = "ACGTACACTGGACTGAGTCT"
        seeds = ["111110000000011111", "111111100001111111"]
        s_seeds = btllib.parse_seeds(seeds)

        seqM1 = "ACGTACACTTGACTGAGTCT"
        seqM2 = "ACGTACACTGTACTGAGTCT"
        seqM3 = "ACGTACACTGCACTGAGTCT"

        k = len(seq) - 2
        self.assertEqual(k, len(seeds[0]))
        self.assertEqual(k, len(seeds[1]))

        seed_nthash = btllib.SeedNtHash(seq, s_seeds, 2, k)
        seed_nthashM1 = btllib.SeedNtHash(seqM1, s_seeds, 2, k)
        seed_nthashM2 = btllib.SeedNtHash(seqM2, s_seeds, 2, k)
        seed_nthashM3 = btllib.SeedNtHash(seqM3, s_seeds, 2, k)

        hashes = []

        self.assertEqual(seed_nthash.get_hash_num(), len(seeds) * 2)

        steps = 0
        while seed_nthash.roll():
            self.assertTrue(seed_nthashM1.roll())
            self.assertTrue(seed_nthashM2.roll())
            self.assertTrue(seed_nthashM3.roll())

            seq_sub = seq[steps:steps + k]
            seqM1_sub = seqM1[steps:steps + k]
            seqM2_sub = seqM2[steps:steps + k]
            seqM3_sub = seqM3[steps:steps + k]

            seed_nthash_base = btllib.SeedNtHash(seq_sub, s_seeds, 2, k)
            seed_nthashM1_base = btllib.SeedNtHash(seqM1_sub, s_seeds, 2, k)
            seed_nthashM2_base = btllib.SeedNtHash(seqM2_sub, s_seeds, 2, k)
            seed_nthashM3_base = btllib.SeedNtHash(seqM3_sub, s_seeds, 2, k)

            self.assertTrue(seed_nthash_base.roll())
            self.assertTrue(seed_nthashM1_base.roll())
            self.assertTrue(seed_nthashM2_base.roll())
            self.assertTrue(seed_nthashM3_base.roll())

            hashes.append([])
            for i in range(seed_nthash.get_hash_num()):
                hval = seed_nthash.hashes()[i]
                self.assertEqual(hval, seed_nthashM1.hashes()[i])
                self.assertEqual(hval, seed_nthashM2.hashes()[i])
                self.assertEqual(hval, seed_nthashM3.hashes()[i])
                self.assertEqual(hval, seed_nthashM1_base.hashes()[i])
                self.assertEqual(hval, seed_nthashM2_base.hashes()[i])
                self.assertEqual(hval, seed_nthashM3_base.hashes()[i])
                hashes[-1].append(hval)

            if seed_nthash.get_pos() > 0:
                seed_nthash.peek_back()
                for i in range(seed_nthash.get_hash_num()):
                    self.assertEqual(seed_nthash.hashes()[i], hashes[-2][i])

                seed_nthash.peek_back(seq[seed_nthash.get_pos() - 1])
                for i in range(seed_nthash.get_hash_num()):
                    self.assertEqual(seed_nthash.hashes()[i], hashes[-2][i])

            steps += 1

        self.assertFalse(seed_nthashM1.roll())
        self.assertFalse(seed_nthashM2.roll())
        self.assertFalse(seed_nthashM3.roll())
        self.assertEqual(steps, len(seq) - k + 1)

    def test_base_substitution(self):
        seq = "ACGTACACTGGACTGAGTCT"
        sub = "ACGCGCACTGGACTGAGTCT"

        nthash = btllib.NtHash(seq, 3, len(seq))
        nthash_subbed = btllib.NtHash(sub, 3, len(sub))

        nthash.roll()
        nthash.sub([3, 4], [ord('C'), ord('G')]) ## convert to ASCII
        nthash_subbed.roll()

        self.assertEqual(nthash.get_hash_num(), nthash_subbed.get_hash_num())
        for i in range(nthash.get_hash_num()):
            self.assertEqual(nthash.hashes()[i], nthash_subbed.hashes()[i])

    def test_kmer_pos(self):
        seq = "CCCTATTAGTACAGTAGTGCCTTCATCGGC"
        h = 3
        k = 6
        nthash = btllib.NtHash(seq, h, k)
        
        for index in range(len(seq) - k):
            nthash.roll()
            self.assertEqual(nthash.get_pos(),index)

    def test_kmer_peeking(self):
        seq = "ACTGATCAG"
        h = 3
        k = 6

        nthash = btllib.NtHash(seq, h, k)
        nthash.roll()

        steps = 3
        while steps:
            nthash.peek()
            h_current = list(nthash.hashes())
            _ = nthash.peek(seq[nthash.get_pos() + k])
            h_peek_next = list(nthash.hashes())
            self.assertEqual(h_current, h_peek_next)
            nthash.roll()
            self.assertEqual(list(nthash.hashes()), h_current)
            steps -= 1



import os
import string
import random
import btllib
import unittest

class SeqWriterTests(unittest.TestCase):

    def test_fasta(self):
        ids = [ "1", "2" ]
        comments = [ "comment1", "comment2" ]
        seqs = [ "ACTG", "TGCA" ]
        quals = [ "!@^&", "(#&$" ]
            
        letters = string.ascii_lowercase
        for iteration in range(3):
            random_filename = ''.join(random.choice(letters) for i in range(64))

            writer_fasta = btllib.SeqWriter(random_filename, btllib.SeqWriter.FASTA)
            for i in range(2):
                writer_fasta.write(ids[i], comments[i], seqs[i], "")
            writer_fasta.close()

            reader_fasta = btllib.SeqReader(random_filename, btllib.SeqReaderFlag.SHORT_MODE)
            self.assertEqual(reader_fasta.get_format(), btllib.SeqReader.SeqReaderFormat_FASTA)

            i = 0
            for record in reader_fasta:
                self.assertEqual(record.id, ids[i])
                self.assertEqual(record.comment, comments[i])
                self.assertEqual(record.seq, seqs[i])
                self.assertTrue(not record.qual)

                i += 1
            self.assertEqual(i, 2)

            reader_fasta.close()
            os.remove(random_filename)
            
    def test_fastq(self):
        ids = [ "1", "2" ]
        comments = [ "comment1", "comment2" ]
        seqs = [ "ACTG", "TGCA" ]
        quals = [ "!@^&", "(#&$" ]
            
        letters = string.ascii_lowercase
        for iteration in range(3):
            random_filename = ''.join(random.choice(letters) for i in range(64))

            writer_fastq = btllib.SeqWriter(random_filename, btllib.SeqWriter.FASTQ)
            for i in range(2):
                writer_fastq.write(ids[i], comments[i], seqs[i], quals[i])
            writer_fastq.close()

            reader_fastq = btllib.SeqReader(random_filename, btllib.SeqReaderFlag.SHORT_MODE)
            self.assertEqual(reader_fastq.get_format(), btllib.SeqReader.SeqReaderFormat_FASTQ)

            i = 0
            for record in reader_fastq:
                self.assertEqual(record.id, ids[i])
                self.assertEqual(record.comment, comments[i])
                self.assertEqual(record.seq, seqs[i])
                self.assertEqual(record.qual, quals[i])

                i += 1
            self.assertEqual(i, 2)

            reader_fastq.close()
            os.remove(random_filename)



import btllib
import unittest


class AAHashTests(unittest.TestCase):

    def test_level_1_aahash(self):
        seq = "RITMLYTI"
        k = 8
        h = 1
        level = 1

        aahash = btllib.AAHash(seq, h, k, level)

        lvl1_hashes = [
            4971716992320218996,
            2124324262552088491,
            8599050776293572050,
            4162540294170059916,
            4246229847304014199,
            3577776033547863068,
            7629354742360239848,
            337326900406420216,
            4971716992320218996
        ]

        num_kmers = len(seq) - k + 2
        idx = 0

        while aahash.roll():
            self.assertEqual(lvl1_hashes[idx], aahash.hashes()[0])
            idx += 1

    def test_level_2_aahash(self):
        seq = "RITMLYTI"
        k = 8
        h = 1
        level = 2

        aahash = btllib.AAHash(seq, h, k, level)

        lvl2_hashes = [
            16984239531076160731,
            2122538798213336131,
            1181634448219258361,
            18364025942200420499,
            15509211063111791705,
            9581438197916650957,
            16088538885945095129,
            6961954824987972818,
            16984239531076160731
        ]

        num_kmers = len(seq) - k + 2
        idx = 0

        while aahash.roll():
            self.assertEqual(lvl2_hashes[idx], aahash.hashes()[0])
            idx += 1

    def test_level_3_aahash(self):
        seq = "RITMLYTI"
        k = 8
        h = 1
        level = 3

        aahash = btllib.AAHash(seq, h, k, level)

        lvl3_hashes = [
            14204475232333016996,
            7411036844653579932,
            4010182074210619839,
            12162032840376646189,
            12359534317481130205,
            12474504577607761213,
            12397957167802074209,
            11500784329128563089,
            14204475232333016996
        ]

        num_kmers = len(seq) - k + 2
        idx = 0

        while aahash.roll():
            self.assertEqual(lvl3_hashes[idx], aahash.hashes()[0])
            idx += 1

    def test_level_2_equivalence(self):
        seq1 = "CGATNDQVWHP"
        seq2 = "CGASNEKIFHP"
        k = 11
        h = 1
        level = 2

        aahash1 = btllib.AAHash(seq1, h, k, level)
        aahash2 = btllib.AAHash(seq2, h, k, level)
        aahash1.roll()
        aahash2.roll()
        self.assertEqual(aahash1.hashes()[0], aahash2.hashes()[0])

    def test_level_3_equivalance(self):
        seq1 = "CGANQVWHP"
        seq2 = "CGTDKIFHP"
        k = 9
        h = 1
        level = 3

        aahash1 = btllib.AAHash(seq1, h, k, level)
        aahash2 = btllib.AAHash(seq2, h, k, level)
        aahash1.roll()
        aahash2.roll()
        self.assertEqual(aahash1.hashes()[0], aahash2.hashes()[0])



import os
import btllib
import unittest

class MIBloomFilterTests(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(__file__)
        
        ### mi_bf_1
        self.test_hashes_1 = [[1, 10, 100], [100, 200, 300]]
        ###########
        
        ### mi_bf_2
        self.random_dna_2 = "GGTAGACACACGTCCACCCCGCTGCTCTGTGACAGGGACTAAAGAGGCGAAGATTATCGTGTGTGCCCCGTTATGGTCGAGTTCGGTCAGAGCGTCATTGCGAGTAGTCGTTTGCTTTCTCGAATTCCGAGCGATTAAGCGTGACAGTCCCAGCGAACCCACAAAACGTGATCGCAGTCCATGCGATCATACGCAAGAAGGAAGGTCCCCATACACCGACGCACCAGTTTACACGCCGTATGCATAAACGAGCTGCACAAACGAGAGTGCTTGAACTGGACCTCTAGTTCCTCTACAAAGAACAGGTTGACCTGTCGCGAAGTTGCCTTGCCTAGATGCAATGTCGGACGTATTACTTTTGCCTCAACGGCTCCTGCTTTCGCTGAAACCCAAGACAGGCAACAGTAACCGCCTTTTGAAGGCGAGTCCTTCGTCTGTGACTAACTGTGCCAAATCGTCTTCCAAACTCCTAATCCAGTTTAACTCACCAAATTATAGCCATACAGACCCTAATTTCATATCATATCACGCCATTAGCCTCTGCTAAAATTCTGTGCTCAAGGGTTTTGGTTCGCCCGAGTGATGTTGCCAATTAGGACCATCAAATGCACATGTTACAGGACTTCTTATAAATACTTTTTTCCTGGGGAGTAGCGGATCTTAATGGATGTTGCCAGCTGGTATGGAAGCTAATAGCGCCGGTGGGAGCGTAATCTGCCGTCTCCACCAACACAACGCTATCGGGTCATATTATAAGATTCCGCAATGGGGTTACTTATAGGTAGCCTTAACGATATCCGGAACTTGCGATGTACGTGCTATGCTTTAATACATACCTGGCCCAGTAGTTTTCCAATATGGGAACATCAATTGTACATCGGGCCGGGATAATCATGTCATCACGGAAGTAGCCGTAAGACAAATAATTCAAAAGAGATGTCGTTTTGCTAGTTCACGTGAAGGTGTCTCGCGCCACCTCTAAGTAAGTGGGCCGTCGAGA"
        self.id_array_2 = [1, 2, 3, 4]
        ###########
        
        ### mi_bf_3
        self.test_hashes_3 = [[1, 10, 100], [100, 200, 300], [500, 1000, 2000]]
        self.id_array_3 = [1, 2, 3, 4] 
        ########### 
                
   
    def set_up_mi_bf_1(self):
        
        self.mi_bf_1 = btllib.MIBloomFilter8(1024 * 1024, 3, "ntHash")
        
        for h in self.test_hashes_1:
            self.mi_bf_1.insert_bv(h)
        self.mi_bf_1.complete_bv_insertion()

    def set_up_mi_bf_2(self):
        
        self.mi_bf_2 = btllib.MIBloomFilter8(1024 * 1024, 3, "ntHash")
        
        nthash = btllib.NtHash(self.random_dna_2, 3, 15)
        while nthash.roll():
            self.mi_bf_2.insert_bv(nthash.hashes())
        self.mi_bf_2.complete_bv_insertion()
        
        for id in self.id_array_2:
            nthash = btllib.NtHash(self.random_dna_2, 3, 15)
            while nthash.roll():
                self.mi_bf_2.insert_id(nthash.hashes(), id)
                
    def set_up_mi_bf_3(self):
        
        self.mi_bf_3 = btllib.MIBloomFilter8(1024 * 1024, 3, "ntHash")
        
        for h in self.test_hashes_3:
            self.mi_bf_3.insert_bv(h)
        self.mi_bf_3.complete_bv_insertion()
        
        for id in self.id_array_3:
            for h in self.test_hashes_3:
                self.mi_bf_3.insert_id(h, id)
        self.mi_bf_3.complete_id_insertion()
        
        for h in self.test_hashes_3:
            for id in self.id_array_3:
                self.mi_bf_3.insert_saturation(h, id)
        
    def test_mibloomfilter_bv_insertion(self):
        
        self.set_up_mi_bf_1() 

        self.assertTrue(self.mi_bf_1.bv_contains([1, 10, 100]))
        self.assertTrue(self.mi_bf_1.bv_contains([100, 200, 300]))
        self.assertFalse(self.mi_bf_1.bv_contains([1, 20, 100]))

    def test_mibloomfilter_id_insertion(self):

        self.set_up_mi_bf_1()
        
        expected_id = 12
        self.mi_bf_1.insert_id(self.test_hashes_1[0], expected_id)

        for result_id in self.mi_bf_1.get_id(self.test_hashes_1[0]):
            self.assertEqual(result_id, expected_id)
            
    def test_mibloomfilter_id_occurence(self):
        self.set_up_mi_bf_1()
        
        expected_id = 12
        self.mi_bf_1.insert_id(self.test_hashes_1[0], expected_id)

        include_saturated = True
        self.assertEqual(len(self.test_hashes_1[0]), 
                         self.mi_bf_1.get_id_occurence_count(include_saturated)[expected_id])

    def test_mibloomfilter_random_sampling(self):
        
        self.set_up_mi_bf_2()

        id_occurences = self.mi_bf_2.get_id_occurence_count(False) ## function returns fixed size large array with zeroes
        total_id_occurences = sum(id_occurences)
        expected_id_occurence = total_id_occurences // len(self.id_array_2) 

        for id in self.id_array_2:
            self.assertAlmostEqual(id_occurences[id], 
                                   expected_id_occurence,
                                   delta = expected_id_occurence // 10)
            
    def test_mibloomfilter_saving(self):
        self.set_up_mi_bf_2()
        
        file_path = os.path.join(self.base_dir, "test.mibf")
        self.mi_bf_2.save(file_path)
       
        try:
            loaded_mi_bf = btllib.MIBloomFilter8(file_path)
        
            expected_id_occurences = self.mi_bf_2.get_id_occurence_count(True)
            id_occurences_loaded = loaded_mi_bf.get_id_occurence_count(True)
            
            for i, val in enumerate(expected_id_occurences):
                self.assertEqual(id_occurences_loaded[i], val)
        finally:
            os.remove(file_path)
            os.remove(file_path+".sdsl")
            
    def test_mibloomfilter_saturation(self):
        
        self.set_up_mi_bf_3()
        
        id_found = {i:False for i in self.id_array_3}
        
        results =  self.mi_bf_3.get_id(self.test_hashes_3[0])
        
        for res in results:
            self.assertTrue(res > self.mi_bf_3.MASK)
            id_found[res & self.mi_bf_3.ANTI_MASK] = True

        self.assertEqual(3,
                         sum(value == True for value in id_found.values()))
